local MAJOR_VERSION = 'LibSidePanel-1.0'
local MINOR_VERSION = '8-SNAPSHOT'

local _G = _G
local LibStub = LibStub
if not LibStub then error(MAJOR_VERSION .. " requires LibStub") end

local oldLib = LibStub:GetLibrary(MAJOR_VERSION, true)
-- Try to create a new library
local sidePanel = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
-- Abort if we are no new library
if not sidePanel then
	return
end

-- we use jostle if it is present, however it is not required
local jostle = LibStub:GetLibrary("LibJostle-3.0", true)

do
	local libDebugLog = LibStub:GetLibrary('LibDebugLog-1.0',true)
	if libDebugLog then
		libDebugLog:Embed(sidePanel)
		sidePanel:ToggleDebugLog(true)
	else
		function sidePanel:Debug (...)
		end
	end
end

-- Preserve our old managed frames or create a new array
-- the content is: { frame, side, priority }
sidePanel.sideFrames = oldLib and oldLib.sideFrames or {}

-- Preserve our old update frame
sidePanel.frame = oldLib and oldLib.frame or CreateFrame('Frame')
local updateFrame = sidePanel.frame

-- Copied from LibJostle-3.0:
-- Start time, used for init stuff
local start = GetTime()
local nextTime = 0

-- Stores whether we have completed our initialization
local fullyInitted = false
updateFrame:SetScript('OnUpdate', function(this, elapsed)
	local now = GetTime()
	if now - start > 0 then
		fullyInitted = true
		-- set the schedule script
		this:SetScript('OnUpdate', function(this, elapsed)
			if GetTime() >= nextTime then
				this:Hide()
				sidePanel:Refresh()
			end
		end)
	end
end)

-- Shedules an update
function updateFrame:Schedule(time)
	time = time or 0
	nextTime = GetTime() + time
	self:Show()
end

local function findFrame(frame)
	for k,v in ipairs(sidePanel.sideFrames) do
		if v.frame and v.frame == frame then
			return v
		end
	end
	return nil
end


--- Registers a frame.
-- @description Adds a frame to be managed, note that the frame must set its height or width (depending on the side) but should not set any other size property.
-- @param frame The frame to register.
-- @param side The side, must bei one of 'TOP','LEFT','BOTTOM', or 'RIGHT'.
-- @param priority Lower priority frames are closer to the edges of the screen (default is 100), must not be negative.
function sidePanel:Register(frame, side, priority)
	-- check for valid side
	if side ~= 'TOP' and side ~= 'LEFT' and side ~= 'BOTTOM' and side ~= 'RIGHT' then
		error("Not a valid side: " .. side)
	end

	-- check if this is a frame
	if (not frame) or (not frame.IsObjectType) or (not frame:IsObjectType('Frame')) then
		error("Not a valid frame")
	end

	if priority < 0 then
		error("Priority must not be negative")
	end

	-- do we already have this frame
	local entry = findFrame(frame)
	local new = false
	if not entry then
		-- new entry
		entry = {}
		new = true
		entry.frame = frame
	else
		assert(entry.frame == frame, "searched for a frame but got a a different")
	end

	entry.side = side
	entry.priority = priority or 100
	-- new entries are inserted, old ones are already there
	if new then
		table.insert(self.sideFrames, entry)
	end
	-- Add to jostle if present
	if jostle then
		if side == 'TOP' then
			jostle:RegisterTop(frame)
		elseif side == 'BOTTOM' then
			jostle:RegisterBottom(frame)
		end
	end

-- 	self:Debug("After insert:")
-- 	for k,entry in ipairs(self.sideFrames) do
-- 		self:Debug(k,entry.frame,entry.side,entry.priority)
-- 	end

	updateFrame:Schedule()
	return
end

--- Unregisters a frame
-- Well also remove the frame from jostle
-- @param frame the frame to unregister
function sidePanel:Unregister(frame)
	for k,v in ipairs(self.sideFrames) do
		if v.frame and v.frame == frame then
			table.remove(self.sideFrames, k)
			break
		end
	end
	-- Remove from jostle if present
	if jostle then
		jostle:Unregister(frame)
	end
	updateFrameSchedule()
end

local function sortFunction(a,b)
	return a.priority < b.priority
end

--- Will reanchor all managed frames.
-- Should not be called directly, as it will be called by Register and Unregister.
function sidePanel:Refresh()
	-- Sort our frames
-- 	self:Debug("Before sorting:")
-- 	for k,entry in ipairs(self.sideFrames) do
-- 		self:Debug(k,entry.frame,entry.side,entry.priority)
-- 	end
 	table.sort(self.sideFrames,sortFunction)
-- 	self:Debug("After sorting:")
-- 	for k,entry in ipairs(self.sideFrames) do
-- 		self:Debug(k,entry.frame,entry.side,entry.priority)
-- 	end

	-- Get UIParent as start border:
	local margin = {}
	margin.tl = { frame = UIParent, point = 'TOPLEFT'     }
	margin.tr = { frame = UIParent, point = 'TOPRIGHT'    }
	margin.bl = { frame = UIParent, point = 'BOTTOMLEFT'  }
	margin.br = { frame = UIParent, point = 'BOTTOMRIGHT' }

	-- Iterate over each entry
	for k,entry in ipairs(self.sideFrames) do
		local side  = entry.side
		local frame = entry.frame
		-- Depending on the side, we need to update our margins
		frame:ClearAllPoints()
		if side == 'TOP' then
			frame:SetPoint( 'TOPLEFT',    margin.tl.frame, margin.tl.point, 0, 1 )
			frame:SetPoint( 'RIGHT',      margin.tr.frame, margin.tr.point, 0, 0 )
			margin.tl.frame = frame
			margin.tl.point = 'BOTTOMLEFT'
			margin.tr.frame = frame
			margin.tr.point = 'BOTTOMRIGHT'
		elseif side == 'LEFT' then
			frame:SetPoint( 'TOPLEFT',    margin.tl.frame, margin.tl.point, 0, 0 )
			frame:SetPoint( 'BOTTOM',     margin.bl.frame, margin.bl.point, 0, 0 )
			margin.tl.frame = frame
			margin.tl.point = 'TOPRIGHT'
			margin.bl.frame = frame
			margin.bl.point = 'BOTTOMRIGHT'
		elseif side == 'BOTTOM' then
			frame:SetPoint( 'BOTTOMLEFT', margin.bl.frame, margin.bl.point, 0, 0 )
			frame:SetPoint( 'RIGHT',      margin.br.frame, margin.br.point, 0, 0 )
			margin.bl.frame = frame
			margin.bl.point = 'TOPLEFT'
			margin.br.frame = frame
			margin.br.point = 'TOPRIGHT'
		elseif side == 'RIGHT' then
			frame:SetPoint( 'TOPRIGHT',   margin.tr.frame, margin.tr.point, 0, 1 )
			frame:SetPoint( 'BOTTOM',     margin.br.frame, margin.br.point, 0, 0 )
			margin.tr.frame = frame
			margin.tr.point = 'TOPLEFT'
			margin.br.frame = frame
			margin.br.point = 'BOTTOMLEFT'
		else
			error("Unexpected side")
		end
	end
end
